# Copyright (C) 2012 Bob Bowles <bobjohnbowles@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from tkinter import *
from tkinter.ttk import *



class ChooseText(Toplevel):
    """Present a text entry field so the user can select the mode of play."""

    def __init__(self, master=None, prompt=None):
        super().__init__(master)
        self.transient(master)
        #self.grid(sticky=(N, S, E, W))

        self.prompt = prompt

        # layout on one row, and two cols
        for c in range(2): self.columnconfigure(c, weight=1, pad=20)
        for r in range(2): self.rowconfigure(r, weight=1, pad=20)

        self.createWidgets()


    def createWidgets(self):
        """Make the entry screen."""
        # instructions/heading
        self.message = Label(self, text=self.prompt, font='helvetica 12 bold')
        self.message.grid(column=0, row=0, columnspan=2, sticky=())

        # set up the entry field
        self.textVar = StringVar(self)
        self.textVar.set('')
        self.text = Entry(self, textvariable=self.textVar)
        self.text.grid(column=0, row=1, sticky=())

        # key binding so we can just hit return to finish typing
        self.text.bind('<Return>', self.hitReturn)

        # hit OK when finished
        self.chosen = BooleanVar(self)
        self.chosen.set(False)
        self.okButton = Checkbutton(self, text='OK',
                                    style='NoIndicator.TCheckbutton',
                                    variable=self.chosen,
                                    onvalue=True, offvalue=False,
                                    command=self.ok
                                    )
        self.okButton.grid(column=1, row=1, sticky=())


    def hitReturn(self, event):
        """Invoke the okButton when the return key is pressed on the entry field."""
        self.okButton.invoke()


    def ok(self):
        """return control to the main program."""
        self.destroy()




# Copyright (C) 2012 Bob Bowles <bobjohnbowles@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import numpy
from reversiConstants import *



class Location(numpy.ndarray):
    """This is a subclass of numpy.ndarray that replaces the need for lists to
    represent [x, y] grid coordinates.
    It includes some handy utility methods for location manipulations. """


    def __new__(cls, location):
        """Boilerplate instantiator for ndarray subclasses."""
        obj = numpy.array(location).view(cls)
        return obj


    def __array_finalize__(self, obj):
        """Boilerplate instance finalizer for ndarray subclasses."""
        if obj is None: return


    def isOnTheBoard(self):
        """Check if the location is on the board."""
        valid = True
        for i in range(2):
            valid = valid and self[i] >= 0 and self[i] < MAX_GRID
        return valid


    def isCorner(self):
        """Check if the location is in a corner of the board. These locations
        are special because they can never be flipped."""
        return self in CORNERS


    def __eq__(self, other):
        """Override parent to avoid ambiguity in equality tests."""
        isEqual = numpy.equal(self, other)
        allEqual = True
        for eq in isEqual: allEqual = allEqual and eq
        return allEqual


    def __str__(self):
        loc = self.tolist()
        return '(' + str(loc[0]+1) + ', ' + str(loc[1]+1) + ')'



# grid direction vectors (8 points of the compass)
DIRECTIONS = []
for x in range(-1, 2):
    for y in range(-1, 2):
        # we skip the centre
        if not (x == y == 0): DIRECTIONS.append(Location([x, y]))

# grid corners
TOP_LEFT = Location([0, 0])
BOTTOM_LEFT = Location([0, MAX_GRID-1])
TOP_RIGHT = Location([MAX_GRID-1, 0])
BOTTOM_RIGHT = Location([MAX_GRID-1, MAX_GRID-1])
CORNERS = [TOP_LEFT, BOTTOM_LEFT, TOP_RIGHT, BOTTOM_RIGHT]




# Copyright (C) 2012 Bob Bowles <bobjohnbowles@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from tkinter import *
from tkinter.ttk import *


def defineStyles(master=None):
    """Define some styles for the display of the grid button icons and
    checkbuttons."""

    style = Style(master)

    # vista theme (default for win7) does not work
    style.theme_use('default')

    # styles for the grid and the scoreboard
    style.configure('Grid.TButton', background='dark green')
    style.configure('Normal.Scoreboard.TButton', background='#dfdfdf')
    style.configure('Selected.Scoreboard.TButton', background='red')

    # adapt the TCheckbutton layout to remove the indicator
    style.layout('NoIndicator.TCheckbutton',
                 [('Checkbutton.padding',
                  {'children':
                   [('Checkbutton.focus',
                     {'children':
                      [('Checkbutton.label', {'sticky': 'nswe'})],
                      'side': 'left', 'sticky': 'nsew'})],
                   'sticky': 'nswe'})])
    # use the TButton map as a template for the new TCheckbutton behaviour
    style.map('NoIndicator.TCheckbutton',
              {'relief': [('!disabled', 'pressed', 'sunken')]})
    # use the TButton configuration as a template for the appearance
    style.configure('NoIndicator.TCheckbutton',
                     padding='3 3',
                     width='-9',
                     shiftrelief=1,
                     relief='raised',
                     anchor='center')






# Copyright (C) 2012 Bob Bowles <bobjohnbowles@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from board import *
from gameControl import *
import random, ui



class Game(object):
    """Encapsulates a game. A game has two players (one human, one computer),
    and a board. The players take turns to make a move.
    The game continues until one of the players wins, or a draw is declared."""


    def __init__(self, players, mode):
        """Initialise the game, and decide who goes first."""

        # initialize the players
        self.players = players
        for player in self.players: player.beginGame()

        # set up the playing mode
        self.mode = mode

        # set up the board
        self.board = Board()

        #set up the board display
        if not self.mode.bench(): ui.displayBoard(self.board)

        # randomly assign who goes first (turn toggles during the game)
        self.turn = random.randint(0,1)


    def play(self):
        """The main loop of the game. Alternate turns between the 2 players
        until a termination condition arises."""
        passed = False    # flag for passes, set if the last move was passed
        while True:

            # decide which player's turn it is
            currentPlayer = self.players[self.turn]

            # flag the player to do something
            if not self.mode.bench(): ui.flagPlayersTurn(currentPlayer)

            try:

                # check there are valid moves remaining
                if not self.board.findAllValidMoves(currentPlayer.token):
                    pass # in this case we catch a GameControl exception anyway
                passed = False

                move = currentPlayer.selectMove(self.board)
            except GameControl as gameControl:
                # if the player requests hints, toggle hints, re-start the move
                if gameControl.instruction == GameControl.HINT:
                    currentPlayer.hints = not currentPlayer.hints
                    flag = _('OFF')
                    if currentPlayer.hints: flag = _('ON')
                    ui.displayMessage(str(_('Hints are %s for %s.') % \
                                          (flag, currentPlayer.name)))
                    continue

                # the player has requested to quit
                if gameControl.instruction == GameControl.QUIT:
                    ui.displayMessage(str(_('%s has quit.') % \
                                          (currentPlayer.name)))
                    break # this is an end-game option!

                # the player has no valid moves
                if gameControl.instruction == GameControl.PASS:
                    if not self.mode.bench():
                        ui.displayMessage(str(_('%s cannot move.') % \
                                              (currentPlayer.name)))
                    if passed: break # the game is over
                    else: passed = True

            if not passed: self.board.makeMove(currentPlayer.token, move)
            self.turn = (self.turn + 1) % 2 # alternate the players

        # print out the final board configuration and scores before exit
        if not self.mode.bench():
            ui.displayBoard(self.board)
            ui.updateGameScores(self.board.getScores())
            ui.displayMessage(_('Game over!'))

        return self.board.getScores()





# Copyright (C) 2012 Bob Bowles <bobjohnbowles@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""the idea of this module is to act as an abstraction layer between the game
and the ui implementation(s)"""

# the ui implementations. Only 1 should be used at a time
#import ui_console as ui_impl
##import ui_pygame as ui_impl # not any more!
import ui_tkinter as ui_impl


def startUI(title, welcomeMessage):
    """Initialize the UI and display a welcome message."""
    ui_impl.startUI(title, welcomeMessage)


def getMode(prompt, modes):
    """The player chooses what mode to play in.
    Returns the mode selected."""
    return ui_impl.getMode(prompt, modes)


def getNumberOfGames(prompt):
    """Obtain the number of games required for a bench test.
    Returns the number."""
    return ui_impl.getNumberOfGames(prompt)


def getPlayersName(prompt):
    """Obtain the player's name.
    Returns what the player types in."""
    return ui_impl.getPlayersName(prompt)


def chooseToken(prompt):
    """The player chooses which token they want to use in the game.
    Returns a key identifying the token (token.tk)."""
    return ui_impl.chooseToken(prompt)


def setupScoreboard(players, score, gamesPlayed, total, wins):
    """Initialise the scoreboard on the UI."""
    ui_impl.setupScoreboard(players, score, gamesPlayed, total, wins)


def displayMessage(message):
    """Give a message to the player."""
    ui_impl.displayMessage(message)


def flagPlayersTurn(player):
    """Indicate to the player whose turn it is."""
    ui_impl.flagPlayersTurn(player)


def displayBoard(board):
    """Display the current state of play on the UI."""
    ui_impl.displayBoard(board)


def chooseMove(player):
    """Player selects a move.
    Returns the selected move, or may raise a GameControl throwable to deal
    with the player deciding to quit or requesting hints."""
    return ui_impl.chooseMove(player)


def playerPause():
    """Wait for the player to do something."""
    ui_impl.playerPause()


def updateGameScores(score):
    """After each move re-calculate and re-display the scores."""
    ui_impl.updateGameScores(score)


def updateGameResults(players, score, gamesPlayed, total, wins):
    """At the end of each game update the players' game statistics."""
    ui_impl.updateGameResults(players, score, gamesPlayed, total, wins)


def getPlayAgain(prompt):
    """Ask the player if they want to play again.
    Return True if they want to play again."""
    return ui_impl.getPlayAgain(prompt)



def stopUI(finishMessage):
    """Display a message and do any final housekeeping on the UI before shutting
    down."""
    ui_impl.stopUI(finishMessage)




# Copyright (C) 2012 Bob Bowles <bobjohnbowles@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.



class GameControl(Exception):
    """This is a customised throwable object to pass control of the game from
    the players to the main game loop with some form of instruction on what to do
    next."""
    # Valid values for instruction:
    HINT = 'hint'
    QUIT = 'quit'
    PASS = 'pass'


    def __init__(self, instruction):
        """Add an attribute that can contain some information."""
        self.instruction = instruction
        Exception.__init__(self)




# Copyright (C) 2012 Bob Bowles <bobjohnbowles@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from tkinter import *
from tkinter.ttk import *

from guiParameters import *
from reversiToken import *
#from board import Board
from location import *
from gridButton import *



class BoardTk(Frame):
    """The playing board holds an 8*8 grid of playing squares to hold the
    playing pieces and for the players to interact with."""


    def __init__(self, master=None):
        Frame.__init__(self, master)
        self.grid(sticky=(N, S, E, W))

        # this attribute is used to detect when a grid button has been clicked
        self.chosen = None

        # set up the 8*8 grid
        for c in range(MAX_GRID): self.columnconfigure(c, weight=1, minsize=75)
        for r in range(MAX_GRID): self.rowconfigure(r, weight=1, minsize=75)

        # now populate the grid
        self.createWidgets()


    def createWidgets(self):
        """Set up the playing board."""
        # use a board grid of tokens as a template data structure
        self.gridArray = self.makeGrid()
        grid = self.gridArray

        # the list of lists of grid locations with their piece of the display
        for x in range(MAX_GRID):
            for y in range(MAX_GRID):

                # initialize the array of ui widgets
                grid[x][y] = GridButton(self, pos=Location([x, y]))
                grid[x][y].grid(column=x, row=y, sticky=(N, S, E, W))


    def makeGrid(self):
        """Return an empty grid data structure."""
        grid = []
        for x in range(MAX_GRID):
            grid.append([None] * MAX_GRID)
        return grid




# Copyright (C) 2012 Bob Bowles <bobjohnbowles@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


# Abstraction layer for the graphics files.

imageDir = 'graphics/'
blackImageFile = imageDir + 'BWSmiley-75x75.gif'
whiteImageFile = imageDir + 'WBSmiley-75x75.gif'
hintImageFile = imageDir + 'star-75x75.gif'
unplayedImageFile = imageDir + 'unplayed-75x75.gif'





#!/usr/bin/env python3
# Copyright (C) 2012 Bob Bowles <bobjohnbowles@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
An implementation of the popular Reversi board game, for use by one or two
players. Some simple AI is included to provide a computer opponent, and if you
are feeling lazy you can watch while two AI opponents slug it out.

The controls for the game are minimal, and hopefully self-explanatory, but here
is a list:

Preparation:
============
Language:   A selection box appears to enable the player to choose their preferred
            language. At time of writing the only options supported are English
            and Chinese, but if there is someone 'out there' interested in
            helping with other translations I will be happy to hear from them.
            If no language is selected the game will default to English.
Your name:  You can type anything you like here, it just helps to distinguish
            the players.
Mode:       Four modes are supported:
    n       Normal (human versus computer opponent).
    p       Person versus Person, in 'hot-seat' mode.
    c       Computer versus computer, what I like to call TV mode.
    b       Benchmarking mode, all graphics turned off. In this mode additional
            information is requested for the number of games to play.
            The default mode is 'Normal'.
Token:      The game asks the first player to choose a token, Black or White.

Game Play:
==========
Who starts is chosen at random.
Hints:      A hint mode is provided for each (human) player that can be toggled
            by either typing 'h' or clicking the appropriate button, depending
            on the UI you are using.
Quit:       Typing 'q', hitting 'Esc' and/or clicking the 'Close Window' icon
            (the details depend on the UI) causes the current game to be
            aborted.
Play Again: You can elect to play again as many times as you want. The more
            games you play, your game statistics will be accumulated and
            displayed on the scoreboard. However, at the time of writing, there
            is no mechanism for storing game stats between sessions.

Bob Bowles <bobjohnbowles@gmail.com>
"""


# initialize the language from a rudimentary UI BEFORE we do any imports
import gettext
import chooseLanguageWindow


from utils import *
from mode import *
from game import *
from player import *
from reversiToken import *
from reversiConstants import *
import ui



# main program and loop


ui.startUI(_('Reversi'), _('Welcome to Reversi!'))

# choose the game mode, set up mode variables
mode = Mode()
mode.debugMode = False   # set the debug flag
totalGames = None
if mode.bench():
    while not totalGames:
        number = ui.getNumberOfGames(_('Enter how many iterations: '))
        try: totalGames = int(number)
        except: continue

# set up the players - dummy 'ghost' player for draws
players = choosePlayers(mode)
draw = Player(DRAW_TK.name(), {NDR:DRAW_TK}, mode)
players.append(draw)

# accumulators for the final totals
gamesPlayed = 0
wins = {NBK: 0, NWH: 0, NDR: 0}
total = {NBK: 0, NWH: 0, NDR: False}

score = {NBK:0, NWH:0}
ui.setupScoreboard(players, score, gamesPlayed, total, wins)

while True:
    # reset the game
    game = Game(players, mode)
    score = game.play()

    # accumulate the totals
    gamesPlayed += 1
    for k in score.keys(): total[k] += score[k]

    # who won?
    if score[NBK] > score[NWH]: wins[NBK] += 1
    elif score[NWH] > score[NBK]: wins[NWH] += 1
    else: wins[NDR] += 1

    # display the scores
    if not mode.bench():
        ui.updateGameResults(players, score, gamesPlayed, total, wins)

    # loop exit
    if mode.bench():
        if gamesPlayed == totalGames: break
    elif not ui.getPlayAgain(_('Play again? (yes or no)')):
            break
    else:
        # reset the scores for the new game
        ui.updateGameScores({NBK:0, NWH:0})

# calculate and display the final results
ui.updateGameResults(players, score, gamesPlayed, total, wins)


ui.stopUI(_('Thank you for playing!'))





# Copyright (C) 2012 Bob Bowles <bobjohnbowles@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from game import *
from board import *
from reversiToken import *
import ui, time



class Player(object):
    """This defines the behaviour and attributes common to all players of the
    game.
    A player is assigned a token.
    When it is the player's turn, the player selects a move and places their
    token on the board."""


    def __init__(self, name, choices, mode):
        """The player has a token that is assigned during instantiation."""
        # initialize a toggle for requesting hints
        self.hints = False
        self.name = name
        self.mode = mode
        self.strategies = [] # this is the plug for computer player AI
        self._chooseToken(choices)


    def beginGame(self):
        """Enables the player to do any necessary preparation at the start of
        the game (reset counters, etc.). This parent method does nothing.
        Implementations of PLayer must provide any required functionality."""
        pass


    def chooseMove(self, board):
        """Empty method to be implemented in subclasses. Returns the player's
        chosen move and the list of tokens to flip as a tuple."""
        return None, None


    def selectMove(self, board):
        """Display the current board and the scores, then select the move for
        the next turn on the board.
        Returns the chosen move and the list of opponent's tokens that will flip
        as a tuple. """
        if not self.mode.bench():
            self.displayBoard(board)
            self.displayScores(board)
        return self.chooseMove(board)


    def _chooseToken(self, choices):
        """The player chooses which token to use at the start of the game.
        The available choices are presented in a list. There are only 2 options.
        The second player has no real choice.
        If the seed letter is initially anything other than 'B' or 'W' a
        dialogue is implemented until a suitable choice has been made."""
        letter = ''
        if len(choices) == 1:
            # the computer (or second player) gets Hobson's choice!
            letter, self.token = choices.popitem()
        else:
            # the human's choice
            letter = ''
            while not letter in choices.keys():
                letter = \
                ui.chooseToken(str(self.name + _(', choose ') + \
                                   _(' or ').join(str('%s' % choices[k].name()) \
                                               for k in choices.keys())))\
                                               .upper()
            self.token = choices.pop(letter)

        if not self.mode.bench() and not self.token == DRAW_TK:
            ui.displayMessage(str(self.name + _(', you play ') + \
                                  self.token.name()))


    def displayBoard(self, board):
        """Display the current board."""
        ui.displayBoard(board)


    def displayScores(self, board):
        """Display the current scores for this player and their opponent."""
        ui.updateGameScores(board.getScores())



class Human(Player):
    """The human subclass of player selects their next move using keyboard input."""


    def chooseMove(self, board):
        """Display the board and accept keyboard input for the human player to
        make their choice of move. Various checks are applied to make sure the
        player types something meaningful to the program.
        Normal exit returns a list with the selected move and the list of tokens
        that are to be flipped.
        Exceptions may be generated by the player that require handling in the
        main game loop."""
        while True:
            move = ui.chooseMove(self)

            # are the coordinates on the board?
            if not move.isOnTheBoard():
                ui.displayMessage(_('Not on board'))
                continue

            # has the position already been taken?
            if not board.isPositionFree(move):
                ui.displayMessage(_('Position occupied'))
                continue

            # is the move valid, i.e., does it result in at least 1 flip?
            tokensToBeFlipped = board.isValidMove(self.token, move)
            if not tokensToBeFlipped:
                ui.displayMessage(\
                        _('Not valid - no flips'))
                continue

            # if the loop gets this far down the move must be OK.
            return move, tokensToBeFlipped


    def displayBoard(self, board):
        """Override the parent method to print hints if they have been
        requested."""
        if self.hints: Player.displayBoard(self, board.giveHints(self.token))
        else: Player.displayBoard(self, board)



class Computer(Player):
    """The computer player uses an AI algorithm to choose its next move."""


    def beginGame(self):
        """The AI player(s) may need to reset some of their strategies where
        these have kept a record of moves in the last game."""
        for strategy in self.strategies: strategy.reset()


    def chooseMove(self, board):
        """Implement the AI logic to choose the computer's next move.
        The computer has a very simple strategy:
        1)     if one of the board corners is a valid move, take it.
        2)     search all possible valid moves and choose the one that gives the
               highest score this turn.
        Returns the chosen move and the flips that will result from it as a
        tuple."""
        # pause here so the human opponent can see the move
        if self.mode.normal():
            ui.playerPause()

        elif self.mode.computer():
            time.sleep(5.0)

        validMoves = board.findAllValidMoves(self.token)
        random.shuffle(validMoves)

        # go through all the strategies this computer has till we find a move
        for strategy in self.strategies:
            move = strategy.findBestMove(self.token, board, validMoves)
            if move: return move

        # provided we have proper fallbacks we should never do this
        return False




# Copyright (C) 2012 Bob Bowles <bobjohnbowles@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from tkinter import *
from tkinter.ttk import *

from boardTk import BoardTk
from scoreBoard import Scoreboard
from reversiToken import *



class GameWindow(Frame):
    """The top-level window for the game"""


    def __init__(self, master=None):
        super().__init__(master)
        self.grid(sticky=(N, S, E, W))

        # define 2 columns and 3 rows with equal weight (for now)
        for c in range(2): self.columnconfigure(c, weight=1)
        for r in range(3): self.rowconfigure(r, weight=1)

        self.createWidgets()


    def createWidgets(self):
        """Define the main components of the game window."""

        # the game board play area
        self.board = BoardTk(self)
        self.board.grid(column=0, row=0, rowspan=3, sticky=(N, S, E, W))

        # title for the scoreboards
        self.title = Label(self, text=_('Scores'), anchor='center',
                           font='helvetica 14 bold')
        self.title.grid(column=1, row=0, sticky=(N, S, E, W))

        # scoreboards for each player
        r = 0
        self.scoreboard = dict.fromkeys(playerTokens)
        for k in self.scoreboard.keys():
            playerScores = Scoreboard(self, k)
            r += 1
            playerScores.grid(column=1, row=r, sticky=(N, S, E, W))
            self.scoreboard[k] = playerScores





# Copyright (C) 2012 Bob Bowles <bobjohnbowles@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""Command-line console implementation of the game UI."""

from gameControl import *
from location import *
from reversiToken import *


def startUI(title, welcomeMessage):
    """Initialize the UI (not needed). Display a welcome message."""
    print(title)
    print()
    print(welcomeMessage)
    print()


def getMode(prompt, modes):
    """The player chooses what mode to play in.
    Returns the mode selected."""
    print(prompt)
    return input('          ' + \
                 '\n          '.join((str('%s: %s' % (k, modes[k])) \
                                      for k in modes.keys())))


def getNumberOfGames(prompt):
    """Obtain the number of required games for a bench test."""
    # TODO: really need a sanity check that the user really has entered a number.
    return int(input(prompt))


def getPlayersName(prompt):
    """Obtain the player's name.
    Returns what the player types in."""
    return input(prompt)


def chooseToken(prompt):
    """The player makes a selection based on the prompt.
    Returns a string. The first letter identifies the choice."""
    return input(prompt)


def setupScoreboard(players, score, gamesPlayed, total, wins):
    """Initialise the scoreboard on the UI. Not required for console."""
    pass


def displayMessage(message):
    """Print a message for the player."""
    print(message)


def flagPlayersTurn(player):
    """Tell the player(s) whose turn it is."""
    print(player.name + _(', your turn!'))


def displayBoard(board):
    """Get the board to print its string representation."""
    print(board)


def chooseMove(player):
    """Interaction between the game and the (human) player. The player is
    required to select their next move. Alternatively they can abort the game,
    or toggle hints.
    Returns the selected location or raises a GameControl throwable to deal with
    hints or quit."""

    move = []
    while not move:
        moveInput = \
        input(_('%s, What is your next move? (x y) x,y = 1-8 quit or hint ')
                 % (player.name)).lower()

        # user hits <return> without typing anything
        if not moveInput: continue

        # user bail-outs are thrown as GameControl instructions.
        if not moveInput[0].isdigit():
            if moveInput.startswith('q'):
                raise GameControl(GameControl.QUIT)
            if moveInput.startswith('h'):
                raise GameControl(GameControl.HINT)
            print(_('Sorry, I do not understand "') + moveInput + _('"'))
            continue

        # make sure player has typed something sensible
        try:
            for i in moveInput.split(): move.append(int(i)-1)
            if len(move) != 2:
                print(_('Please type two numbers'))
                continue
            return Location(move)
        except TypeError:
            print(_('Move must have numeric coordinates'))
            continue


def playerPause():
    """Wait for the player to hit the <enter> key."""
    input(_('Press <enter> when ready'))


def updateGameScores(score):
    """After each move re-calculate and re-display the scores."""
    print(_('Scores:  ') + _(' vs. ').join(str('%s: %2r'
          % ((k, score[k]))) for k in score.keys()))


def updateGameResults(players, score, gamesPlayed, total, wins):
    """At the end of each game update the players' game statistics."""
    print(_('Results for %3r games:') % gamesPlayed)
    print(_('Averages:                       Scores:       Wins:'))
    for player in players:
        k = str(player.token)
        if total[k]: average = str('%2.2d' % (total[k]/gamesPlayed))
        else: average = '    '
        print('               %15s %4s: %4s        %2r (%2.1d%%)' % \
              (player.name, k, average, wins[k], (wins[k]/gamesPlayed*100.0)))


def getPlayAgain(prompt):
    """Ask the player if they want to play again.
    Return True if they want to play again."""
    return input(prompt).startswith('y')


def stopUI(finishMessage):
    """Just say goodbye."""
    print(finishMessage)




# Copyright (C) 2012 Bob Bowles <bobjohnbowles@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


# constants for handling tokens
NBK = 'B'
NWH = 'W'
NH = 'H'
NU = 'U'
NDR = 'D'
tokenPlayNames = {NBK:_('Black'), NWH:_('White')} # for the players
tokenScoreNames = {NBK:_('Black'),
                   NWH:_('White'),
                   NDR:_('Drawn')} # for scoring
tokenNames = {NBK:_('Black'),
              NWH:_('White'),
              NH:_('Hint'),
              NU:_('Unplayed'),
              NDR:_('Drawn')}

# These constants are specific to console output
TB = '\u263b'
TW = '\u263a'
TH = '.'
TU = ' '
printTk = {NBK:TB, NWH:TW, NH:TH, NU:TU} # these are for console use



class Token(object):
    """Defines attributes of a player token."""


    def __init__(self, tk=NU):
        """Initialize the token's state. The default is 'unplayed'."""
        self.tk = tk


    def __str__(self):
        """Enable the token to print itself."""
        return self.tk


    def __eq__(self, other):
        """Token comparison."""
        return self.tk == other.tk


    def name(self):
        """Calculate the token's name from its state."""
        return tokenNames[self.tk]


    def play(self, token):
        """Player plays this token. The token may be another instance of Token,
        or just a letter."""
        if self.__class__ == token.__class__:
            self.tk = token.tk
        else: self.tk = token


    def flip(self):
        """Flip the token from one player to the other."""
        if self.tk == NBK: self.tk = NWH
        else: self.tk = NBK


    def mark(self):
        """Change the token to display itself as playable (hint mode)."""
        self.tk = NH


    def isEmpty(self):
        """Confirm if the token has not been played."""
        return self.tk == NU


    def clone(self):
        """Make a copy."""
        return Token(self.tk)


    def setTk(self, tk):
        """A setter to help disambiguate 'tk' when using tkinter."""
        self.tk = tk



# Constant token types for reference
BLACK_TK = Token(NBK)
WHITE_TK = Token(NWH)
DRAW_TK = Token(NDR)
UNPLAYED_TK = Token(NU)
HINT_TK = Token(NH)

playerTokens = {NBK:BLACK_TK, NWH:WHITE_TK}




# Copyright (C) 2012 Bob Bowles <bobjohnbowles@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import ui


class Mode(object):
    """The mode of play. Valid modes are:
    (p)vp:            'hot-seat' human vs. human
    (n)ormal:         human vs. computer
    (c)omputer:       computer vs. computer (TV mode)
    (b)enchtest:      computer vs. computer (benchtest - no visuals)"""
    PVP = 'p'
    NORMAL = 'n'
    COMPUTER = 'c'
    BENCH = 'b'
    MODES = {NORMAL: _('Normal human vs. computer'),
             PVP: _('Person vs. person'),
             COMPUTER: _('Computer vs. computer TV mode'),
             BENCH: _('Bench-test (no UI)')}


    def __init__(self, mode=None):
        """Can define this mode during construction, or initiate a user dialog."""
        if not mode or mode not in Mode.MODES.keys(): mode = self.chooseMode()
        self.mode = mode
        self.debugMode = False # toggle for debugging mode


    def chooseMode(self):
        """Set up mode of play interactively."""
        modes = Mode.MODES      # shorthand
        mode = ''
        while mode not in modes.keys():
            mode = ui.getMode(_('Choose a mode:'), modes)
        return mode


    def normal(self):
        """Returns whether the mode is normal."""
        return self.mode == Mode.NORMAL

    def pvp(self):
        """Returns whether the mode is PVP."""
        return self.mode == Mode.PVP

    def computer(self):
        """Returns whether the mode is computer vs. computer."""
        return self.mode == Mode.COMPUTER

    def bench(self):
        """Returns whether the mode is bench-test."""
        return self.mode == Mode.BENCH

    def debug(self):
        """Returns whether the mode is debug."""
        return self.debugMode

    def __str__(self):
        return 'Mode is ' + self.mode + ': ' + self.MODES[self.mode] + \
            ' Debug: ' + str(self.debug())




# Copyright (C) 2012 Bob Bowles <bobjohnbowles@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from strategy import *
from player import *
from reversiToken import *
from mode import *



# Utility functions for the main loop

def giveName():
    """Get the player's name."""
    return ui.getPlayersName(_('Your name:'))


def choosePlayers(mode):
    """Set up the players depending on the mode of play.
    Returns a list containing the two players."""
    # Computer strategy options
    VANILLA_STRATEGY = [FindACorner(), MaximizeScore()]
    CLEVER_STRATEGY = [FindUnflippable(), MaximizeScore()]


    # allow the first player to choose their token (if it is human)
    choices = playerTokens.copy()

    # set up player 1
    if mode.normal() or mode.pvp(): # player 1 is human
        player1 = Human(giveName(), choices, mode)
    else:       # the other modes use a vanilla computer for player 1
        player1 = Computer(_('Mr Niceguy'), {NBK:BLACK_TK}, mode)
        player1.strategies = VANILLA_STRATEGY

    # set up player 2
    if mode.pvp():    # player 2 is human only for pvp mode
        player2 = Human(giveName(), choices, mode)
    elif mode.normal():   # token choice is different for the computers
        player2 = Computer(_('Mr Cleverclogs'), choices, mode)
        player2.strategies = CLEVER_STRATEGY # 'clever'
    else:
        player2 = Computer(_('Mr Cleverclogs'), {NWH:WHITE_TK}, mode)
        player2.strategies = CLEVER_STRATEGY # 'clever'

    return [player1, player2]




# Copyright (C) 2012 Bob Bowles <bobjohnbowles@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from tkinter import *
from tkinter.ttk import *

from reversiToken import *
from gridButton import *



class Scoreboard(Frame):
    """Panel to hold the scores for a player."""


    def __init__(self, master=None, key=NBK):
        super().__init__(master)
        self.grid(sticky=(N, S, E, W))
        self.configure(padding=20)
        self.key = key

        # set up a 3*6 grid
        for c in range(2): self.columnconfigure(c, weight=1)
        for r in range(7): self.rowconfigure(r, weight=1)

        # now populate the grid
        self.createWidgets()


    def createWidgets(self):
        """Define 6 data display objects for the player, showing:
        0) the player's token
        1) the player's name
        2) their current score
        3) the number of wins
        4) percentage wins
        5) average score"""

        # player's token goes at the top
        self.icon = GridButton(self,
                               token=self.key,
                               style='Scoreboard.TButton')
        self.icon.grid(column=0, row=0, columnspan=2, sticky=())

        # player's name
        self.nameVal = StringVar(self)
        self.nameVal.set('Name')
        self.name = Label(self, textvariable=self.nameVal, anchor='center',
                          font='helvetica 12 bold')
        self.name.grid(column=0, row=1, columnspan=2, sticky=(E, W))

        # current score
        self.scoreLbl = Label(self, text=_('Score:'))
        self.scoreLbl.grid(column=0, row=2, sticky=(E, W))
        self.scoreVal = IntVar()
        self.score = Label(self, textvariable=self.scoreVal, justify='right')
        self.score.grid(column=1, row=2, sticky=(E))

        # number of wins
        self.winsLbl = Label(self, text=_('Wins:'))
        self.winsLbl.grid(column=0, row=3, sticky=(E, W))
        self.winsVal = IntVar()
        self.wins = Label(self, textvariable=self.winsVal, justify='right')
        self.wins.grid(column=1, row=3, sticky=(E))

        # Percentage
        self.percentLbl = Label(self, text='%:')
        self.percentLbl.grid(column=0, row=4, sticky=(E, W))
        self.percentVal = IntVar()
        self.percent = Label(self, textvariable=self.percentVal,
                             justify='right')
        self.percent.grid(column=1, row=4, sticky=(E))

        # average score
        self.averageLbl = Label(self, text=_('Average:'))
        self.averageLbl.grid(column=0, row=5, sticky=(E, W))
        self.averageVal = IntVar()
        self.average = Label(self, textvariable=self.averageVal,
                             justify='right')
        self.average.grid(column=1, row=5, sticky=(E))

        # button to handle hints
        self.hints = BooleanVar(self)
        self.hints.set(False)
        self.hintsButton = Checkbutton(self, text=_('Hints'),
                                       style='NoIndicator.TCheckbutton',
                                       variable=self.hints,
                                       onvalue=True, offvalue=False)
        self.hintsButton.grid(column=1, row=6, sticky=(E))





# Copyright (C) 2012 Bob Bowles <bobjohnbowles@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""Tkinter implementation of the game GUI."""
from tkinter import *
from tkinter.ttk import *
from tkinter import messagebox

from gameControl import *
from location import *
from reversiToken import *
from reversiConstants import *
from board import *
from player import *

from gameWindow import *
from chooseModeWindow import *
from chooseNumberWindow import *
from chooseTextWindow import *
from chooseTokenWindow import *
import styles



#==============================================================================
# UI Callbacks


# callback for the main window so we can kill it cleanly.
quit = False
def quitGame():
    """Exit the current game."""
    global quit
    quit = True


#==============================================================================
# UI INTERFACE IMPLEMENTATION


def startUI(title, welcomeMessage):
    """Initialize the UI. Display a welcome message."""
    global rootWindow, gameWindow

    # start the main window and set up window-killing callback
    rootWindow = Tk()
    rootWindow.resizable(True, True)
    rootWindow.protocol('WM_DELETE_WINDOW', quitGame)

    # implement styles
    styles.defineStyles(rootWindow)

    # put up a title
    rootWindow.title(title)

    # the main game window with the board and score panel
    gameWindow = GameWindow(rootWindow)
    rootWindow.update()

    # simple welcome message
    messagebox.showinfo(message=welcomeMessage)


def getMode(prompt, modes):
    """The player chooses what mode to play in.
    Returns the mode selected."""
    chooseModeWindow = ChooseMode(rootWindow, prompt, modes)

    mode = None
    while not chooseModeWindow.chosen.get():
        mode = chooseModeWindow.radioVar.get()
        chooseModeWindow.update()
    return mode


def getNumberOfGames(prompt):
    """Obtain the number of required games for a bench test."""
    chooseNumberWindow = ChooseNumber(rootWindow, prompt)
    numberOfGames = 0
    while not chooseNumberWindow.chosen.get():
        numberOfGames = chooseNumberWindow.numberVar.get()
        chooseNumberWindow.update()
    return numberOfGames


def getPlayersName(prompt):
    """Obtain the player's name.
    Returns what the player types in."""
    chooseTextWindow = ChooseText(rootWindow, prompt)
    name = ''
    while not chooseTextWindow.chosen.get():
        name = chooseTextWindow.textVar.get()
        chooseTextWindow.update()
    return name


def chooseToken(prompt):
    """The player makes a selection based on the prompt.
    Returns a string. The first letter identifies the choice."""
    chooseTokenWindow = ChooseToken(rootWindow, prompt)
    tokenID = ''
    while not tokenID:
        tokenID = chooseTokenWindow.tokenVar.get()
        chooseTokenWindow.update()
    return tokenID


def findPlayer(players, key):
    """Identify a player from their playing token.
    Returns the player with the matching token key."""
    for player in players:
        if player.token.tk == key: return player


def setupScoreboard(players, score, gamesPlayed, total, wins):
    """Initialise the scoreboard on the UI. Note all parameters are dictionaries
    indexed on token parameter."""
    for k in score.keys():
        scoreboard = gameWindow.scoreboard[k]

        # set up the players' names
        player = findPlayer(players, k)
        scoreboard.nameVal.set(player.name)

        # initialise the scores
        scoreboard.scoreVal.set(score[k])

        # set up wins
        scoreboard.winsVal.set(wins[k])

        # set up percentages
        percent = 0
        if gamesPlayed: percent = wins[k] * 100 // gamesPlayed
        scoreboard.percentVal.set(percent)

        # set up averages
        average = 0
        if gamesPlayed: average = total[k] // gamesPlayed
        scoreboard.averageVal.set(average)

        # remove the hints button for non-human players
        if isinstance(player, Computer): scoreboard.hintsButton.forget()

    gameWindow.update()


def displayMessage(message):
    """Print a message for the player."""
    gameWindow.update()
    messagebox.showinfo(message=message)
    gameWindow.update()


def flagPlayersTurn(player):
    """Tell the player(s) whose turn it is by changing the background of the
    icons on the scoreboard. Make sure the correct hint button is enabled."""
    gameWindow.update()
    turn = player.token.tk
    for k in gameWindow.scoreboard.keys():

        # set the icon to show which player's move is next
        icon = gameWindow.scoreboard[k].icon
        if k == turn: icon.configure(style='Selected.Scoreboard.TButton')
        else: icon.configure(style='Normal.Scoreboard.TButton')

        # enable/disable the hints button for human players if applicable
        if k == turn and isinstance(player, Human):
            gameWindow.scoreboard[k].hintsButton.configure(state='normal')
        else:
            gameWindow.scoreboard[k].hintsButton.configure(state='disabled')

    gameWindow.update()


def displayBoard(board):
    """Map the current state of the board onto the GUI."""
    gameWindow.update()
    for x in range(MAX_GRID):
        for y in range(MAX_GRID):
            token = board.grid[x][y]
            guiButton = gameWindow.board.gridArray[x][y]
            guiButton.token.set(token.tk)
    gameWindow.update()


def chooseMove(player):
    """Interaction between the game and the (human) player. The player is
    required to select their next move. Alternatively they can abort the game,
    or toggle hints.
    Returns the selected location or raises a GameControl throwable to deal with
    hints or quit."""

    move = None
    while move == None:
        gameWindow.update()

        # check to see if the window quit button was clicked
        global quit
        if quit:
            quit = False
            raise GameControl(GameControl.QUIT)

        # check the scoreboard to see if a hints button state has changed
        if gameWindow.scoreboard[player.token.tk].hints.get() != player.hints:
            raise GameControl(GameControl.HINT)

        # check the gui grid to see if anything has been clicked
        move = gameWindow.board.chosen

    # clear the grid selection latch
    gameWindow.board.chosen = None
    gameWindow.update()

    return move


def playerPause():
    """Wait for the player to clear the messagebox. Actually this plays slicker
    without the messagebox."""
    gameWindow.update()
    # TODO maybe have a menu option to toggle this
    #messagebox.showinfo(message=_('Click <OK> when ready...'))
    #gameWindow.update()


def updateGameScores(score):
    """After each move re-calculate and re-display the scores."""
    for k in score.keys():
        gameWindow.scoreboard[k].scoreVal.set(score[k])

    gameWindow.update()


def updateGameResults(players, score, gamesPlayed, total, wins):
    """At the end of each game update the players' game statistics."""
    for k in score.keys():
        scoreboard = gameWindow.scoreboard[k]

        # update wins
        scoreboard.winsVal.set(wins[k])

        # update percentages
        percent = 0
        if gamesPlayed: percent = wins[k] * 100 // gamesPlayed
        scoreboard.percentVal.set(percent)

        # update averages
        average = 0
        if gamesPlayed: average = total[k] // gamesPlayed
        scoreboard.averageVal.set(average)

    gameWindow.update()


def getPlayAgain(prompt):
    """Ask the player if they want to play again.
    Return True if they want to play again."""
    if messagebox.askyesno(message=prompt): return True
    else: return False


def stopUI(finishMessage):
    """Say goodbye and shut down the UI cleanly."""
    messagebox.showinfo(message=finishMessage)
    gameWindow.update()
    gameWindow.destroy()





# Copyright (C) 2012 Bob Bowles <bobjohnbowles@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from tkinter import *
from tkinter.ttk import *

from reversiToken import *
from location import *

# the image file names - also used in the pygame interface
from graphics import *

# place-holders. We have to set up the images at run-time.
#blackImage = 'BWSmiley-75x75.gif'
#whiteImage = 'WBSmiley-75x75.gif'
#hintImage = 'star-75x75.gif'
#unplayedImage = 'unplayed-75x75.gif'
tokenImageChoices = {NBK: blackImageFile, NWH: whiteImageFile} # for player choice
tokenImage = {NBK: blackImageFile,
              NWH: whiteImageFile,
              NH: hintImageFile,
              NU: unplayedImageFile} # for board play



class TokenVar(StringVar):
    """Subclass of StringVar enables us to set the image and token attributes of
    the widget by game interaction (I hope)."""


    def __init__(self, master=None, value=NU, name=None):
        """Make sure we have references to the parent widget. Use the overridden
        set method to callout to the parent to set up the graphics."""
        super().__init__(master, value, name)
        self.set(value)


    def set(self, token):
        """Extend the parent to set the image and token attributes in the
        parent."""
        super().set(token)
        self._master.setToken(token)



class GridButton(Button):
    """Visual representation of a Reversi board grid square. Contains all the
    information it needs to display itself and interact with the game when
    clicked."""


    def __init__(self, master=None,
                 token=NU, pos=TOP_LEFT, style='Grid.TButton'):
        """Do the class initialization and prepare the game-specific attributes."""
        super().__init__(master, text=str(token), style=style)
        self.master = master

        # this button's grid position and its current token.
#        self.token = token
        self.pos = pos

        # set up the image
#        self.setToken(token.tk)
        self.token = TokenVar(self, token)
        self['textvariable'] = self.token

        # set up the callout
        self['command'] = self.selected


    def setToken(self, key):
        """TokenVar callout to change the displayed image to match the token."""
#        self.token.tk = key
        file = tokenImage[key]
        image = PhotoImage(master= self.master, file=file)
        self.image = image
        self.configure(image=image)


    def selected(self):
        """This is the callout for when the grid location is chosen by a player.
        The location is presented to the parent board, where it can be inspected
        by the game loop."""
        self.master.chosen = self.pos




# Copyright (C) 2012 Bob Bowles <bobjohnbowles@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from tkinter import *
from tkinter.ttk import *



class ChooseMode(Toplevel):
    """Present a radiobutton so the user can select the mode of play."""

    def __init__(self, master=None, prompt=None, modes=None):
        super().__init__(master)
        self.transient(master)
        self.prompt = prompt
        self.modes = modes

        # layout on one col, and however many rows it takes...
        for c in range(1): self.columnconfigure(c, weight=1, pad=20)
        for r in range(1): self.rowconfigure(r, weight=1, pad=20)
        r = 0
        for k in self.modes.keys():
            r += 1
            self.rowconfigure(r, weight=1, pad=5)
        self.rowconfigure(r+1, weight=1, pad=20)

        self.createWidgets()


    def createWidgets(self):
        """Make the radio button choice screen."""
        # instructions/heading
        self.message = Label(self, text=self.prompt, font='helvetica 12 bold')
        self.message.grid(column=0, row=0, sticky=())

        # set up the choices on the radio buttons
        r = 0
        self.radioVar = StringVar(self)
        for k in self.modes.keys():
            r += 1
            Radiobutton(self, text=self.modes[k],
                        variable=self.radioVar, value=k)\
                        .grid(column=0, row=r, sticky=(W), padx=20)
        self.radioVar.set('n') # TODO: hardwire here - can we soften it?

        # hit OK when finished
        self.chosen = BooleanVar(self)
        self.chosen.set(False)
        self.okButton = Checkbutton(self, text='OK',
                                    style='NoIndicator.TCheckbutton',
                                    variable=self.chosen,
                                    onvalue=True, offvalue=False,
                                    command=self.ok
                                    )
        self.okButton.grid(column=0, row=r+1, sticky=(E))


    def ok(self):
        """return control to the main program."""
        self.destroy()




# Copyright (C) 2012 Bob Bowles <bobjohnbowles@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from mode import *
from reversiToken import *
from location import *
from gameControl import *



class Board(object):
    """The board manages a record of the currently completed moves in the game
    on a grid."""


    def __init__(self):
        """
Construct a new copy of the grid with blanks in all the locations except the
starting positions. The starting map should look like this:
  1 2 3 4 5 6 7 8
 +-+-+-+-+-+-+-+-+
1| | | | | | | | |
 +-+-+-+-+-+-+-+-+
2| | | | | | | | |
 +-+-+-+-+-+-+-+-+
3| | | | | | | | |
 +-+-+-+-+-+-+-+-+
4| | | |B|W| | | |
 +-+-+-+-+-+-+-+-+
5| | | |W|B| | | |
 +-+-+-+-+-+-+-+-+
6| | | | | | | | |
 +-+-+-+-+-+-+-+-+
7| | | | | | | | |
 +-+-+-+-+-+-+-+-+
8| | | | | | | | |
 +-+-+-+-+-+-+-+-+

NB The grid indices go from 0 to 7 but the displayed locations go from 1 to 8."""

        # make the grid
        grid = self.makeGrid()

        # populate the grid with the starting positions
        grid[3][3] = BLACK_TK.clone()
        grid[3][4] = WHITE_TK.clone()
        grid[4][3] = WHITE_TK.clone()
        grid[4][4] = BLACK_TK.clone()

        self.grid = grid


    def makeGrid(self):
        """Return blank grid."""
        grid = []
        for x in range(MAX_GRID):
            grid.append([UNPLAYED_TK.clone()] * MAX_GRID)
        return grid


    def __str__(self):
        """Create a printable representation of the board. This implementation
        adds the location indices to make life easier for the human players."""

        # the header lines
        HLINE = ' +-+-+-+-+-+-+-+-+\n'
        HNUM =  '  1 2 3 4 5 6 7 8\n'

        # the display string
        display = '\n'
        display += HNUM
        display += HLINE
        for y in range(MAX_GRID):
            display += str(y+1) + '|'
            for x in range(MAX_GRID):
                display += printTk[self.grid[x][y].tk] + '|'
            display += '\n'
            display += HLINE
        return display


    def clone(self):
        """Make a deep copy of the board and return it.
        We make a new board and assign a clone of this one's grid to it."""
        copy = []
        for x in self.grid:
            copyX = []
            for y in x:
                copyX.append(y.clone()) # clone the tokens at each grid position
            copy.append(copyX)

        clone = Board()
        clone.grid = copy
        return clone


    def isPositionFree(self, move):
        """Determine if the selected move is available."""
        return self.grid[move[0]][move[1]].isEmpty()


    def isValidMove(self, token, move):
        """Check that the move location represents a valid move on the board.
        To satisfy this, the move must result in flipping (reversing) at least
        one of the opponent's tokens. To test this, we do the move as a trial
        and keep track of the flips that can be done.
        Returns the list of tokens to be flipped if true, otherwise False."""

        # shorthand
        g = self.grid
        tk = token                  # the current player's token
        ot = tk.clone()             # the other player's token (the other token)
        ot.flip()
        tokensToFlip = []

        # work out which directions have flippable opponent's tokens.
        for vector in DIRECTIONS:
            loc = move + vector

            # check if we have found an adjacent other token
            if loc.isOnTheBoard() and g[loc[0]][loc[1]] == ot:

                # keep going in the same direction...
                loc += vector

                # ...unless we go off the edge of the board...
                if not loc.isOnTheBoard(): continue

                # ...until we find one of our own tokens, or a blank
                while g[loc[0]][loc[1]] == ot:
                    loc += vector

                    # if we go over the edge break the while and continue the if
                    if not loc.isOnTheBoard(): break
                if not loc.isOnTheBoard(): continue

                # have we found one of our own tokens? We can flip some ot's!
                if g[loc[0]][loc[1]] == tk:
                    # record the locations of the flippable ot's for later
                    while True:
                        loc -= vector
                        if loc == move: break
                        tokensToFlip.append(loc.copy())

        # the move is not valid if there are no flippable ot's
        if len(tokensToFlip) == 0: return False

        return tokensToFlip


    def giveHints(self, token):
        """Request hints from the board for valid move options."""
        hints = self.clone()

        # find the valid options, mark them and provide them in the clone.
        hints.markValidMoves(self.findAllValidMoves(token))
        return hints


    def markValidMoves(self, validMoves):
        """Mark the listed moves on the board."""
        for move in validMoves:
            self.grid[move[0]][move[1]].mark()


    def findAllValidMoves(self,token):
        """Do an exhaustive search of the board for valid moves. Returns the
        list of valid options. If none are found we throw a GameControl exception
        to announce having lost."""
        validMoves = []
        for x in range(MAX_GRID):
            for y in range(MAX_GRID):
                move = Location([x, y])

                # we are only interested in the empty squares
                if not self.isPositionFree(move): continue

                # check if this move is valid, add to the list if so
                if not self.isValidMove(token, move): continue
                validMoves.append(move)

        # if there are no valid moves for this player they have lost the game
        if len(validMoves) == 0: raise GameControl(GameControl.PASS)

        return validMoves


    def makeMove(self, token, moveInstructions):
        """Make the move on the board with the specified token."""
        # unpack the move instructions
        move = moveInstructions[0]
        tokensToBeFlipped = moveInstructions[1]

        # put the new token on the board
        self.grid[move[0]][move[1]] = token.clone()

        # flip the affected tokens of the opponent's
        for otherToken in tokensToBeFlipped:
            self.grid[otherToken[0]][otherToken[1]].flip()


    def getScores(self):
        """Evaluate the current scores of both players by counting their tokens."""
        score = {NBK: 0, NWH: 0}
        for x in self.grid:
            for y in x:
                key = str(y)
                if key in score: score[key] += 1
        return score




# Copyright (C) 2012 Bob Bowles <bobjohnbowles@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Work around mbcs bug in distutils. (for wininst)
# http://bugs.python.org/issue10945
import codecs
try:
    codecs.lookup('mbcs')
except LookupError:
    ascii = codecs.lookup('ascii')
    func = lambda name, enc = ascii: {True: enc}.get(name == 'mbcs')
    codecs.register(func)

from distutils.core import setup
import os

# references to file trees...

# the data files for the app (files NOT located in the package root)
dataFiles = []

# automate the list of i18n files

# locale root
targetLocaleRoot = ''
if os.name == 'nt':
    # need to point to the correct root file for the python version
    import sys
    targetLocaleRoot = os.path.dirname(os.path.realpath(sys.executable))
elif os.name == 'posix': targetLocaleRoot = os.path.join('/', 'usr', 'share')

# auto-generate the list of .mo files
languages = os.listdir('locale')
languageFiles = []
for language in languages:
    languageFiles.append((os.path.join(targetLocaleRoot,
                                       'locale',
                                       language,
                                       'LC_MESSAGES'),
                          [os.path.join('locale',
                                        language,
                                        'LC_MESSAGES',
                                        'Reversi.mo'), ]))
dataFiles.extend(languageFiles)

# detect posix to decide what to do with the launcher
if os.name == 'posix':
    print('Posix detected! Adding desktop launcher.')
    dataFiles.append((os.path.join(targetLocaleRoot, 'applications'),
                      ['Reversi.desktop']))

    # this code is redundant, retained just-in-case.
#     from subprocess import Popen, PIPE
#     pipe = Popen('ps aux | grep unity', shell=True, stdout=PIPE).stdout
#
#     # now scan the output for some unity tell-tale
#     if  'unity-panel-service' in str(pipe.read()):
#         print('Ubuntu and Unity detected!')
#         dataFiles.append((os.path.join(targetLocaleRoot, 'applications'),
#                           ['Reversi.desktop']))
#     else:
#         # TODO: no launcher for other posix
#         print('Posix detected - No Unity')

elif os.name == 'nt':
    # make the bat file install in a vaguely sensible place
    print('Windows detected')

    # NOTE documents and settings still exists on win7, its just hidden.
    publicDesktop = \
        os.path.realpath('C:/Documents and Settings/All Users/Desktop')

    # discriminate between XP and Win7
    if os.path.exists(os.path.realpath('C:/Users')):
        print('Windows 7/8 detected')
        publicDesktop = os.path.realpath('C:/Users/Public/Desktop')
    else:
        print('Windows XP assumed')
    dataFiles.append((publicDesktop, ['Reversi.bat']))

else:
    # no os identified, let's cross our fingers ...
    print('OS not known')

# sort out package data (e.g. gifs etc used in the app)
fileList = []
packageDir = 'src'
packageRoot = os.path.join(packageDir, 'Reversi')

# the images are under the package root
imageRoot = 'graphics'
for file in os.listdir(os.path.join(packageRoot, imageRoot)):
    fileList.append(os.path.join(imageRoot, file))

packageData = {'Reversi': fileList}

# get a reference to the version number from the package being built
import sys
sys.path.insert(0, packageDir)
from Reversi import __version__


# now do the setup
setup(
    name='Reversi',
    version=__version__,
    description='A version of the Reversi board game intended for casual play.',
    long_description=open('README.txt').read(),
    author='Bob Bowles',
    author_email='bobjohnbowles@gmail.com',
    url='http://pypi.python.org/pypi/Reversi',
    license='GNU General Public License v3 (GPLv3)',  # TODO belt-n-braces??
    keywords=[
              "Reversi",
              "Othello",
              ],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Development Status :: 5 - Production/Stable",
        "Environment :: Other Environment",
        "Natural Language :: Chinese (Simplified)",
        "Natural Language :: English",
        "Natural Language :: French",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Microsoft :: Windows :: Windows 7",
        "Topic :: Games/Entertainment :: Board Games",
        "Topic :: Games/Entertainment :: Turn Based Strategy",
        ],
    package_dir={'': packageDir},
    packages=['Reversi', ],
    requires=[  # TODO needs Python >3.3
              'numpy (>=1.6.1)',
              'tkinter (>=8.5)'
              ],
    package_data=packageData,
    data_files=dataFiles,
)






# Copyright (C) 2012 Bob Bowles <bobjohnbowles@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from tkinter import *
from tkinter.ttk import *

from gridButton import *
from reversiToken import *



class ChooseToken(Toplevel):
    """Present a number entry field so the user can select the mode of play."""

    def __init__(self, master=None, prompt=None):
        super().__init__(master)
        self.transient(master)
        # self.grid(sticky=(N, S, E, W))
        self.prompt = prompt

        # layout on two rows, and two cols
        for c in range(2): self.columnconfigure(c, weight=1, pad=20)
        for r in range(1): self.rowconfigure(r, weight=1, pad=20)

        self.createWidgets()


    def createWidgets(self):
        """Make the entry screen."""
        # instructions/heading
        self.message = Label(self, text=self.prompt, font='helvetica 12 bold')
        self.message.grid(column=0, row=0, columnspan=2, sticky=())

        # set up the variable to contain the result
        self.tokenVar = StringVar(self)
        self.tokenVar.set('')

        # display the buttons for the two choices
        c = 0
        for k in playerTokens.keys():
            gridButton = GridButton(self,
                                    token=playerTokens[k].tk,
                                    style='Scoreboard.TButton')
            gridButton.grid(column=c, row=1, sticky=())
            gridButton.configure(pad=20,
                                 command=lambda arg1=gridButton:
                                 self.choose(arg1))
            c += 1


    def choose(self, gridButton):
        """make a selection and return control to the main program."""
        self.tokenVar.set(gridButton.token.get())
        self.destroy()




# Copyright (C) 2012 Bob Bowles <bobjohnbowles@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from mode import *
from location import *
from reversiToken import *



class Strategy(object):
    """This is an abstract class that defines an interface for AI strategies to
    implement in subclasses."""

    def findBestMove(self, token, board, validMoves):
        """Find the best move for the current board, given the current player's
        token, the board, and a list of valid moves.
        Returns EITHER a couplet containing the best move, and the list of
        locations of the opponent's tokens that it will flip,
                OR False if no move is found."""
        return False


    def reset(self):
        """Reset any internal counters before starting a new game."""
        pass



class MaximizeScore(Strategy):
    """This is the fallback strategy if all other strategies fail to find a
    move. It should always be the last strategy implemented. """

    def findBestMove(self, token, board, validMoves):
        """Searches the list of valid moves for one that maximises the score for
        the player's next move.
        Returns a couplet containing the best move, and the list of associated
        flippable opponent's tokens.
        NB Unlike other strategy findBestMove() methods, this method ALWAYS
        returns a move, unless the list of valid moves is incorrectly configured
        or empty."""
        scoreHWM = 0
        for trialMove in validMoves:
            tokensToFlip = board.isValidMove(token, trialMove)

            # do the move on a cloned board and work out the score
            trialBoard = board.clone()
            trialBoard.makeMove(token, (trialMove, tokensToFlip))
            score = trialBoard.getScores()[str(token)]

            # keep the move if the score is a new high and reset the HWM
            if score > scoreHWM:
                scoreHWM = score
                bestMove = trialMove
                bestFlips = tokensToFlip

        return bestMove, bestFlips



class FindACorner(Strategy):
    """This strategy relies on the fact that the corners of the board can never
    be flipped. This makes the corners special."""


    def findBestMove(self, token, board, validMoves):
        """Finds a corner move and chooses it.
        Returns the corner move and the list of opponent's tokens it will flip,
                or False if no corner move is found."""
        for trialMove in validMoves:
            if trialMove.isCorner():
                return trialMove, board.isValidMove(token, trialMove)
        return False



class FindUnflippable(FindACorner):
    """This is a development of the 'find a corner' strategy. Finding a corner
    works because the token cannot be flipped. This strategy looks for corners
    and other positions where the player's token cannot be flipped.
    To simplify the calculation this strategy maintains a list of existing known
    stable locations for the player this strategy belongs to."""


    def __init__(self):
        """Initialize an internal list of locations for existing stable tokens."""
        self.stableTokenLocations = []


    def reset(self):
        """Reset the record of stable tokens at the start of each game."""
        self.__init__()


    def findBestMove(self, token, board, validMoves):
        """This strategy does the following and returns the first move it finds:
        1) Look for a corner move (using parent method).
        2) Look for any move next to at least 4 other unflippable tokens or
        positions off the board.
        (This uses and maintains an internal record of known stable tokens).
        Returns the move and the list of flippable opponent's tokens,
                or False if no move is found."""
#        if mode.debug(): self._dbgShowList('Starting')

        # 1) find a corner
        cornerFlips = FindACorner.findBestMove(self, token, board, validMoves)
        if cornerFlips:
            self.stableTokenLocations.append(cornerFlips[0])
#            if mode.debug(): print('Stable corner found: ', cornerFlips[0])

            # see what other tokens this move makes stable on a trial board
            testBoard = board.clone()
            testBoard.makeMove(token, cornerFlips)
            self._checkNeighbours(token, testBoard, cornerFlips[0])
#            if mode.debug(): self._dbgShowList('Corner')
            return cornerFlips

        # 2) find the first unflippable move in the list
        for move in validMoves:
            flips = board.isValidMove(token, move)

            # see what tokens this move makes stable on a trial board
            testBoard = board.clone()
            testBoard.makeMove(token, (move, flips))
            if self._isStablePosition(token, testBoard, move):
                self._checkNeighbours(token, testBoard, move)
#                if mode.debug(): self._dbgShowList('Unflippable')
                return move, flips

        # fallback if no move is found
        return False


    def _dbgShowList(self, title):
        """DBG: Debug print output of the current list of stable tokens."""
        print(title + ' list of stables: \n' + \
              '[' + ', '.join(str(x) for x in self.stableTokenLocations) + ']')


    def _checkNeighbours(self, token, board, stable):
        """When a location has been verified stable and accepted as a good move,
        some neighbouring tokens may become stable. We re-run the stability
        check, but this time on same-token neighbours.
        For each neighbour we find is stable, we recursively check their
        neighbours, too."""
        # check the neighbours
#        if mode.debug(): print('Checking neighbours of ', stable)
        for direction in DIRECTIONS:
            neighbour = stable + direction

            # we ignore positions not on the board
            if not neighbour.isOnTheBoard():
#                if mode.debug():
#                    print('Neighbour of ', stable, ' is not on the board: ', \
#                          neighbour)
                continue

            # if the position is empty it is irrelevant
            elif board.isPositionFree(neighbour):
#                if mode.debug():
#                    print('Neighbour of ', stable, ' is empty: ', neighbour)
                continue

            # we may already have recorded it
            elif neighbour in self.stableTokenLocations:
#                if mode.debug():
#                    print('Neighbour of ', stable, ' already noted: ', \
#                          neighbour)
                continue

            # if it is a same token make sure it is stable...
            elif board.grid[neighbour[0]][neighbour[1]] == token:
                if self._isStablePosition(token, board, neighbour):
#                    if mode.debug():
#                        print('Neighbour of ', stable, \
#                              ' now identified stable: ', neighbour)
                    self._checkNeighbours(token, board, neighbour)
#                elif mode.debug():
#                    print('Neighbour of ', stable, ' NOT stable: ', neighbour)
#            elif mode.debug():
#                print('Neighbour of ', stable, ' is the opponents!: ', \
#                      neighbour)


    def _isStablePosition(self, token, board, location):
        """Check if the given location with the given token is stable.
        We define 'stable' as having a board edge or an existing stable token
        in at least 4 directions, or filled board locations in all 8 directions
        to the board edge.
        Returns True if the token can not be flipped in that location."""
        # check the 8 neighbours of the location - are they stable?
        stableDirs = []

        # only need do this if our list of known stables is non-empty
        if len(self.stableTokenLocations) != 0:
            for direction in DIRECTIONS:
                nbor = location + direction

                # is the neighbour off the board?
                if not nbor.isOnTheBoard(): stableDirs.append(direction)

                # is the neighbour vacant?
                elif board.isPositionFree(nbor): continue

                # is the neighbour one of ours?
                elif board.grid[nbor[0]][nbor[1]] != token: continue

                # is the neighbour already in our list of stable tokens?
                elif nbor in self.stableTokenLocations:
                    stableDirs.append(direction)

                # is the neighbour a to-be-identified stable?
                else:
                    # see if we find a stable anchor in this direction
                    nbor += direction
                    while nbor.isOnTheBoard() and \
                    board.grid[nbor[0]][nbor[1]] == token:
                        if nbor in self.stableTokenLocations:
                            stableDirs.append(direction)
                            break
                        else:
                            nbor += direction

        # we need to check contiguity of directions for n=4
        contiguousNeighbours = True
        if len(stableDirs) == 4:
            sumSepSq = 0.0                      # sum of squares of separations
            dirsToCheck = stableDirs[:]
            for dir1 in stableDirs:
                dirsToCheck.remove(dir1)
                for dir2 in dirsToCheck:
                    separation = dir2 - dir1
                    sumSepSq += separation.dot(separation) # numpy dot product

            # if sumSepSq is > 14 the directions are not contiguous on the board
            contiguousNeighbours = sumSepSq <= 14
#            if mode.debug():
#                print('Contiguity is ' + str(contiguousNeighbours) + \
#                      ' for ' + str(location))

        if contiguousNeighbours and len(stableDirs) >= 4:
#            if mode.debug(): print('Stable location found: ', location)

            # update the list of stables
            self.stableTokenLocations.append(location)
            return True

        # secondary check to see if all 8 directions are full
        else: # i.e. the last check failed
            for direction in DIRECTIONS:
                nbor = location + direction
                if not nbor.isOnTheBoard(): continue
                while not board.grid[nbor[0]][nbor[1]].isEmpty():
                    nbor += direction
                    if not nbor.isOnTheBoard(): break
                return False

        # we have failed to find a False condition so it must be stable
#        if mode.debug():
#            print('Secondary check found stable location: ', location)
        self.stableTokenLocations.append(location) # update the list
        return True




# Copyright (C) 2012 Bob Bowles <bobjohnbowles@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module abstracts the languages that are implemented in the Reversi
package. When a new language translation is available, the .mo file is inserted
in the appropriate sub-directory, indexed by language_REGION tags. This file is
edited to add the language to the languageIds dictionary. The languageIds
defined here are used to pull in the translations at run time.

The entries in the languageIds dictionary are of the form
    {language_REGION: description},
where the description is *in an appropriate language*.

NB: It is therefore important that these text strings are **NOT** translated by
gettext."""

import gettext, os, sys

# this is the list of available translations. NB NOT TRANSLATED
languageIds = {'en_GB': 'English',
               'fr_FR': 'Français',
               'zh_CN': '中文',
               }  # NB NOT TRANSLATED

# the domain for languages
domain = 'Reversi'

# set the locale directory according to installation
#     choose {python_install_path}\locale for windows
#            /usr/share/locale for Posix
#             ../../locale if neither of the above work (development default)
localedir = os.path.join('..', '..', 'locale')  # development location is default
testPath = ''
if os.name == 'nt':  # windows
    testPath = os.path.join(os.path.dirname(os.path.realpath(sys.executable)),
                            'locale')
elif os.name == 'posix':  # everything else
    testPath = os.path.join('/', 'usr', 'share', 'locale')
# print('languages: Test locale path is ' + testPath)  # test only

# a crude test of the language directory tree.
if os.path.exists(os.path.join(testPath,
                               'en_GB',
                               'LC_MESSAGES',
                               'Reversi.mo')): localedir = testPath
# print('languages: Selected locale directory is ' + localedir)  # test only

# make the translations
languages = dict.fromkeys(languageIds)
for language in languageIds.keys():
    languages[language] = gettext.translation(domain,
                                              localedir=localedir,
                                              languages=[language],
                                              )

# set up a default language in case none gets picked
gettext.install(domain, localedir)
